﻿using System;
using System.Collections.ObjectModel;
using System.ComponentModel;

// ReSharper disable InconsistentNaming

namespace IdeaBlade.EntityModel {

  public abstract class CoroutineOperationWrapper<TEventArgs> 
    : CoroutineOperationWrapper where TEventArgs : CoroutineCompletedEventArgsWrapper {

    protected CoroutineOperationWrapper(TEventArgs completedEventArgs)
      : base(completedEventArgs) {}

    protected new TEventArgs CompletedEventArgs {
      get { return (TEventArgs) base.CompletedEventArgs; }
    }

    private event EventHandler<TEventArgs> _completed;
    public event EventHandler<TEventArgs> Completed {
      add {
        if (IsCompleted) {
          value(this, CompletedEventArgs);
          return;
        }
        _completed += value;
      }
      remove { _completed -= value; }
    }

    protected override void RaiseCompleted(object sender, CoroutineCompletedEventArgs args) {
      CallCompletedActions();
      var handlers = _completed;
      _completed = null;
      if (null == handlers) return;
      handlers(this, CompletedEventArgs);
    }
  }

  public abstract class CoroutineOperationWrapper : INotifyPropertyChanged, INotifyCompleted {

    protected CoroutineOperationWrapper(CoroutineCompletedEventArgsWrapper completedEventArgs) {

      if (null == completedEventArgs) {   
        throw new ArgumentNullException("completedEventArgs");
      }
      CompletedEventArgs = completedEventArgs;
      TriggerOperation = completedEventArgs.TriggerOperation;
      TriggerOperation.Completed += RaiseCompleted;
      TriggerOperation.PropertyChanged += (s, args) => RaisePropertyChanged(args.PropertyName);
    }

    protected CoroutineCompletedEventArgsWrapper CompletedEventArgs { get; private set; }
    protected CoroutineOperation TriggerOperation { get; private set; }
    private BaseOperation BaseOperation { get { return TriggerOperation; } }

    protected abstract void RaiseCompleted(object sender, CoroutineCompletedEventArgs args);

    public string TargetOperationName {
      get { return CompletedEventArgs.TargetOperationName; }
      set { CompletedEventArgs.TargetOperationName = value; }
    }

    public ReadOnlyCollection<INotifyCompleted> Notifications {
      get { return TriggerOperation.Notifications; }
    }

    public bool CanCancel { get { return BaseOperation.CanCancel; } }
    public void Cancel() { BaseOperation.Cancel(); }
    public bool Cancelled { get { return BaseOperation.Cancelled; } }
    public bool CompletedSuccessfully { get { return BaseOperation.CompletedSuccessfully; } }
    public bool CompletedSynchronously { get { return BaseOperation.CompletedSynchronously; } }
    public Exception Error { get { return BaseOperation.Error; } }
    public bool HasError { get { return BaseOperation.HasError; } }
    public bool IsCompleted { get { return BaseOperation.IsCompleted; } }
    public bool IsErrorHandled {
      get { return BaseOperation.IsErrorHandled; }
      set { BaseOperation.IsErrorHandled = value; }
    }
    public void MarkErrorAsHandled() { BaseOperation.MarkErrorAsHandled(); }
    public object UserState { get { return BaseOperation.UserState; } }

    public event PropertyChangedEventHandler PropertyChanged = delegate { };

    protected void RaisePropertyChanged(string propertyName) {
      PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
    }

    protected void CallCompletedActions() {
      var actions = _notifyCompletedActions;
      _notifyCompletedActions = null;
      if (null == actions) return;
      actions(CompletedEventArgs);
    }

    private Action<INotifyCompletedArgs> _notifyCompletedActions;
    public virtual void WhenCompleted(Action<INotifyCompletedArgs> completedAction) {
      if (null == completedAction) return;
      if (IsCompleted) {
        completedAction(CompletedEventArgs);
        return;
      }
      _notifyCompletedActions = 
        (Action<INotifyCompletedArgs>)Delegate.Combine(_notifyCompletedActions, completedAction);
    }

  }

}