// (c) Copyright IdeaBlade, Inc. and Microsoft Corporation
// Derived from Microsoft Silverlight User Test Framework
// This source is subject to the Microsoft Public License (Ms-PL).
// Please see http://go.microsoft.com/fwlink/?LinkID=131993 for details.
// All other rights reserved.

using System;
using System.Linq;
using System.ComponentModel;
using Microsoft.Silverlight.Testing;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace IdeaBlade.TestFramework {
  /// <summary>
  /// Custom test method runner class providing the ability 
  /// to perform semi-asynchronous test tasks. 
  /// </summary>
  /// <remarks>
  /// Modeled on the "WorkItemTest" class in Silverlight Unit Test Framework.
  /// Relies upon a "Dispatcher" equivalent, the
  /// <see cref="TestSynchronizationContext"/>
  /// for executing queued tasks on a common thread.
  /// </remarks>
  public class AsyncTestMethodRunner {

    public AsyncTestMethodRunner() {
      WorkItemContainer = new WorkItemContainer();
      WorkItemContainer.UnhandledException += HandleUnhandledException;
      
      _syncContext = new TestSynchronizationContext();
      _syncContext.UnhandledException += HandleUnhandledException;
      IsRunning = true;
    }

    private readonly TestSynchronizationContext _syncContext;
    private Exception _unhandledException;
    private readonly object _locker = new object();

    private void HandleUnhandledException(object sender, UnhandledExceptionEventArgs e) {
      lock (_locker) {
        if (null != _unhandledException) return; // only care about first exception
        _syncContext.ClearQueue();
        _unhandledException = (e.ExceptionObject as Exception) ?? new Exception("Unknown exception thrown in WorkItem");
        TestComplete();
      }
    }

    protected bool IsRunning { get; set;}


    /// <summary>
    /// Wait for <see cref="TestComplete"/> signal that 
    /// queued async tasks are done.
    /// Typically the last line of the test method.
    /// </summary>
    /// <remarks>
    /// Perpetually processes the queue of tasks until hears the signal to stop. 
    /// Various <see cref="WaitFor"/> methods add tasks to the queue.
    /// <para>
    /// Once <see cref="WaitForTasksToFinish"/> is called, 
    /// the test method will hang until it hears the signal or the test times out.
    /// The signal is almost always the result of calling  <see cref="TestComplete"/>.
    /// </para><para>
    /// Exception catcher must raise <see cref="TestComplete"/> also
    /// to stop the loop and handle the exception.
    /// </para>
    /// </remarks>
    protected internal void WaitForTasksToFinish() {

      while (IsRunning && WorkItemContainer.Invoke()) {}

      _syncContext.ClearQueue(); // just in case  
      ReportResult();
    }

    private void ReportResult()  {
      var bugAtty = GetTestMethodBugAttribute();
      if (null == bugAtty || bugAtty.Fixed) {
        if (null != _unhandledException) { throw _unhandledException; }
      } else {
        var prefix = String.Format(
          "Test marked as a Bug: '{0}'", bugAtty.Description);
        if (null == _unhandledException) {
          Assert.Fail(prefix + "; expected test failure but didn't fail. Should mark it fixed?");
        } else {
          Console.WriteLine(prefix + " and failed as expected.");
          Console.WriteLine(_unhandledException);
        }
      }
    }

    private BugAttribute GetTestMethodBugAttribute() {
      var callStack = new System.Diagnostics.StackTrace();
      // Start looking for BugAttribute on test method
      // starting 4 frames up (count from 0)
      for (var i = 3; i < callStack.FrameCount; i++) {
        var frame = callStack.GetFrame(i); 
        var method = frame.GetMethod();
        var bugAtty = method.GetCustomAttributes(typeof(BugAttribute), false)
            .FirstOrDefault();
        if (null != bugAtty) return (BugAttribute) bugAtty;
      }
      return null;
    }

    /// <summary>
    /// Gets the current test task container.
    /// </summary>
    /// <returns>The current container for the test's tasks.</returns>
    [EditorBrowsable(EditorBrowsableState.Never)]
    protected CompositeWorkItem WorkItemContainer { get; private set; }

    /// <summary>
    /// Signal that a test is complete when using Async testing.  
    /// 
    /// WARNING: If you use your own methods for completing, such as an 
    /// HtmlTimer or other threading method, it is possible that this call 
    /// will occur *AFTER* the test has timed out when using Timeouts.  As 
    /// such, be very careful as you could complete the call to the *next*  test.
    /// </summary>
    public virtual void TestComplete() {
      lock (_locker) {
        _syncContext.ClearQueue();
        WorkItemContainer.WorkItemCompleteInternal();
      }
    }

    /// <summary>
    /// Get thread id of <see cref="System.Threading.SynchronizationContext"/>
    /// used by this test.
    /// </summary>
    public int SyncContextThreadId { get { return _syncContext.SyncContextThreadId; } }

    /// <summary>
    /// Add a task object to the test queue.  For a test that is currently 
    /// executing, all tasks contained within the queue are executed to 
    /// completion (unless an Exception is thrown) -before- moving on to 
    /// the next test.
    /// </summary>
    /// <param name="testTaskObject">Asynchronous test task instance.</param>
    public virtual void EnqueueWorkItem(WorkItem testTaskObject) {
        WorkItemContainer.Enqueue(testTaskObject);
    }

    /// <summary>
    /// Delay a minimum amount of time before continuing. Similar to a sleep
    /// call that is not a blocking call.
    /// </summary>
    /// <param name="delay">The minimum time span to wait before continuing.</param>
    public virtual void EnqueueDelay(TimeSpan delay) {
      EnqueueWorkItem(CreateDelay(delay));
    }

    /// <summary>
    /// Creates a delay work item instance.
    /// </summary>
    /// <param name="delay">The minimum time span to wait before continuing.</param>
    /// <returns>Returns a new work item.</returns>
    [EditorBrowsable(EditorBrowsableState.Never)]
    public WorkItem CreateDelay(TimeSpan delay) {
      return new SleepWorkItem(delay);
    }

    /// <summary>
    /// Delay a minimum amount of time before continuing. Similar to a sleep
    /// call that is not a blocking call.
    /// </summary>
    /// <param name="milliseconds">The minimum number of milliseconds to wait
    /// until the delay is finished.</param>
    public virtual void EnqueueDelay(double milliseconds) {
      EnqueueDelay(TimeSpan.FromMilliseconds(milliseconds));
    }

    /// <summary>
    /// Requires a bool returning delegate to be passed in. Instructs the 
    /// test task queue to wait until the conditional call returns True to 
    /// continue executing other test tasks and/or ending the test method.
    /// </summary>
    /// <param name="conditionalDelegate">Conditional method or delegate. 
    /// Test will halt until this condition returns True.</param>
    public virtual void EnqueueConditional(Func<bool> conditionalDelegate) {
      EnqueueWorkItem(CreateConditional(conditionalDelegate));
    }

    /// <summary>
    /// Creates a conditional work item instance.
    /// </summary>
    /// <param name="conditionalDelegate">Conditional method or delegate. 
    /// Test will halt until this condition returns True.</param>
    /// <returns>Returns a new work item.</returns>
    [EditorBrowsable(EditorBrowsableState.Never)]
    public WorkItem CreateConditional(Func<bool> conditionalDelegate) {
      return new ConditionalWorkItem(conditionalDelegate, _syncContext);
    }

    /// <summary>
    /// Enqueue a test task which calls the TestComplete method of 
    /// SilverlightTest.
    /// </summary>
    public virtual void EnqueueTestComplete() {
      // Don't do with SyncContext
      EnqueueWorkItem(new CallbackWorkItem(TestComplete));
      //EnqueueCallback(TestComplete); // was
    }

    /// <summary>
    /// Add a Callback method into the test task queue.  Similar to the 
    /// PumpMessages(...) call, with the difference being that there is no 
    /// longer a single requirement: you can enqueue several callback 
    /// methods and other test tasks, all of which will execute before the 
    /// test completes and/or the engine continues.
    /// </summary>
    /// <param name="testCallbackDelegate">Void-returning delegate, 
    /// anonymous delegates work fine too.</param>
    public virtual void EnqueueCallback(Action testCallbackDelegate) {
      EnqueueWorkItem(new CallbackWorkItem(testCallbackDelegate, _syncContext));
    }

    /* WB: No usage found */
    ///// <summary>
    ///// Creates a callback work item instance.
    ///// </summary>
    ///// <param name="testCallbackDelegate">Void-returning delegate, 
    ///// anonymous delegates work fine too.</param>
    ///// <returns>Returns a new work item.</returns>
    //[EditorBrowsable(EditorBrowsableState.Never)]
    //public WorkItem CreateCallback(Action testCallbackDelegate) {
    //  return new CallbackWorkItem(testCallbackDelegate);
    //}

    /// <summary>
    /// Adds a number of callback methods into the test task queue.
    /// </summary>
    /// <param name="actions">Set of Action instances.</param>
    public void EnqueueCallback(params Action[] actions) {
      if (actions != null) {
        foreach (Action action in actions) {
          EnqueueCallback(action);
        }
      }
    }

    /// <summary>
    /// Process an exception using the test engine logic for 
    /// ExpectedExceptions and logging the exception if needed.
    /// </summary>
    /// <param name="ex">Exception object.</param>
    public virtual void HandleException(Exception ex) {
      WorkItemContainer.WorkItemException(ex);
    }

  }

  internal class WorkItemContainer : CompositeWorkItem {
    public WorkItemContainer() {
      FinishWhenEmpty = false;
    }
  }
}

// Placeholder namespace to prevent compile failures
// when test files are linked to both Desktop and Silverlight projects.
// ReSharper disable EmptyNamespace
namespace Microsoft.Silverlight.Testing {}
// ReSharper restore EmptyNamespace