using System;
using System.Collections.Generic;
using System.Security.Permissions;
using System.Threading;
using System.Diagnostics;

// ReSharper disable RedundantNameQualifier

namespace IdeaBlade.TestFramework {

  /// <summary>
  /// A Faked <see cref="SynchronizationContext"/> for Desktop Async Tests.
  /// </summary>
  /// <remarks>
  /// This synchronization context implementation comes from Brian Noyes'
  /// Prism-for-Windows-Forms implementation
  /// and is based on the ThreadPoolSynchronizer sample
  /// from IDesign Inc, www.idesign.net.
  /// Brian emailed it to Ward on 15 Aug 2010.
  /// <para>
  /// Added handling of unhandled exceptions thrown in WorkItem.
  /// Changed name from SynchronizationContextMock
  /// and namespace from IDesign.Composite.Testing.Mocks.
  /// Changed "WorkItem" class to "SyncWorkItem" because conflicted
  /// with my "WorkItem" class in this namespace.
  /// "Cleaned-up" to mostly "conform" to IdeaBlade coding standards.
  /// Added some sugar (extra ctor, simplifying overloads of Send/Post).
  /// </para>
  /// </remarks> 
  [SecurityPermission(SecurityAction.Demand, ControlThread = true)]
  public class TestSynchronizationContext : SynchronizationContext, IDisposable {

    public const string TestSynchronizationContextTheadName =
      "TestSynchronizationContext";

    private readonly WorkerThread _workerThread;
    private readonly Queue<SyncWorkItem> _syncWorkItemQueue;

    [ThreadStatic]
    private int _userStateCounter ;

    public TestSynchronizationContext() : this(TestSynchronizationContextTheadName) {}

    public TestSynchronizationContext(string contextName) {
      CallQueued = new Semaphore(0, Int32.MaxValue);
      _syncWorkItemQueue = new Queue<SyncWorkItem>();
      _workerThread = new WorkerThread(contextName, this);
    }

    public int SyncContextThreadId {
      get { return _workerThread.ThreadObj.ManagedThreadId; }
    }

    protected Semaphore CallQueued { get; private set; }

    internal virtual void QueueWorkItem(SyncWorkItem workItem) {
      lock (_syncWorkItemQueue) {
        _syncWorkItemQueue.Enqueue(workItem);
        CallQueued.Release();
      }
    }

    protected virtual bool QueueEmpty {
      get {
        lock (_syncWorkItemQueue) {
          return _syncWorkItemQueue.Count <= 0;
        }
      }
    }

    internal virtual SyncWorkItem GetNext() {
      CallQueued.WaitOne();
      lock (_syncWorkItemQueue) {
        return _syncWorkItemQueue.Count == 0 ? null : _syncWorkItemQueue.Dequeue();
      }
    }


    public override SynchronizationContext CreateCopy() { return this; }

    public void Post(Action act) {
      Post(state => act(), _userStateCounter++);
    }

    public override void Post(SendOrPostCallback method, object state) {
      var workItem = new SyncWorkItem(method, state);
      QueueWorkItem(workItem);
    }

    public void Send(Action act) {
      Send( state => act(), _userStateCounter++);
    }

    public override void Send(SendOrPostCallback method, object state) {
      if (SynchronizationContext.Current == this) {
        //If already on the correct context, must invoke now to avoid deadlock
        ExecuteNow(method, state); 
        // method(state); //this original line wouldn't manage exceptions consistently
        return;
      }
      var workItem = new SyncWorkItem(method, state);
      QueueWorkItem(workItem);
      workItem.Wait();
    }

    private void ExecuteNow(SendOrPostCallback method, object state) {
      try {
        method(state);
      } catch (Exception ex) {
        _workerThread.RaiseUnhandledException(ex);
      }
    }

    public void ClearQueue() {
      lock (_syncWorkItemQueue) {
        _syncWorkItemQueue.Clear();
      } 
    }

    public event EventHandler<UnhandledExceptionEventArgs> UnhandledException {
      add { _workerThread.UnhandledException   += value;  }
      remove { _workerThread.UnhandledException -= value;  }
    }

    public void Dispose() {
      Close();
    }

    public void Close() {
      if (CallQueued.SafeWaitHandle.IsClosed) {
        return;
      }
      CallQueued.Release(Int32.MaxValue);
      _workerThread.Kill();
      _workerThread.ClearUnhandledExceptionHandlers();
      CallQueued.Close();
    }

    public void Abort() {
      CallQueued.Release(Int32.MaxValue);
      _workerThread.ThreadObj.Abort();
      _workerThread.ClearUnhandledExceptionHandlers();
      CallQueued.Close();
    }

    #region WorkThread

    internal class WorkerThread {

      public readonly Thread ThreadObj;
      private readonly TestSynchronizationContext _syncContext;
      private bool _endLoop;

      internal WorkerThread(string name, TestSynchronizationContext context) {
        _syncContext = context;

        _endLoop = false;

        ThreadObj = new Thread(Run) {IsBackground = true, Name = name};
        ThreadObj.Start();
      }

      internal bool EndLoop {
        set { lock (this) { _endLoop = value; } }
        get { lock (this) { return _endLoop; } }
      }

      internal void Run() {
        SetSynchronizationContext(_syncContext);

        while (EndLoop == false) {
          ExecuteWorkItem(_syncContext.GetNext());
        }
      }

      internal void ExecuteWorkItem (SyncWorkItem workItem) {
        if (null == workItem) return;
        Exception ex = workItem.CallBack();
        if (null != ex) RaiseUnhandledException(ex);
      }

      #region Exception handling

      internal event EventHandler<UnhandledExceptionEventArgs> UnhandledException {
        add { lock (this) { _unhandledException += value; } }
        remove { lock (this) { _unhandledException -= value; } }
      }

      // ReSharper disable InconsistentNaming
      private event EventHandler<UnhandledExceptionEventArgs> _unhandledException;
      // ReSharper restore InconsistentNaming
       
      internal void RaiseUnhandledException(Exception ex) {
        OnUnhandledException(new UnhandledExceptionEventArgs(ex, false));
      }

      protected virtual void OnUnhandledException(UnhandledExceptionEventArgs e) {
        EventHandler<UnhandledExceptionEventArgs> handler;
        lock (this) { handler = _unhandledException; }
        if (null == handler) return;
        handler(this, e);
      }

      internal void ClearUnhandledExceptionHandlers() {
        lock (this) {_unhandledException = null;}
      }

      #endregion

      public void Kill() {
        //Kill is called on client thread - must use cached thread object
        Debug.Assert(ThreadObj != null);
        if (ThreadObj.IsAlive == false) {
          return;
        }
        EndLoop = true;

        //Wait for thread to die
        ThreadObj.Join();
      }
    }

    #endregion

  }

  [Serializable]
  internal class SyncWorkItem {

    private readonly object _userState;
    private readonly SendOrPostCallback _method;
    private readonly ManualResetEventSlim _asyncWaitHandle;

    internal void Wait() {
      _asyncWaitHandle.Wait();
    }

    internal SyncWorkItem(SendOrPostCallback method, object state) {
      _method = method;
      _userState = state;
      _asyncWaitHandle = new ManualResetEventSlim(false);
    }

    // Don't execute; just terminate
    internal void Skip() {
      _asyncWaitHandle.Set();
    }

    //This method is called on the worker thread to execute the method
    internal Exception CallBack() {
      Exception except = null;
      try {
        _method(_userState);
      } catch (Exception ex) {
        except = ex;
      }
      _asyncWaitHandle.Set();
      return except;
    }
  }


}
