using System;
using System.Collections.Generic;
using IdeaBlade.EntityModel;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace IdeaBlade.TestFramework {

  public static class OperationTestForErrorExtensions  {

    public static TestTheResultsWrapper<T> TestForError<T>(
      this EntityQueryOperation<T> op, params Action<Exception>[] onErrorActions) {

      op.WireCompletedWithErrorActions(onErrorActions);
      return new TestTheResultsWrapper<T>(op);
    }

    public static TestTheResultsWrapper<T> TestForError<T>(
      this EntityQueryOperation<T> op, string failureMessage = null) {

      WaitFor.IsCompleted(op);
      WaitFor.Callback(() => AssertNoQueryFailure(op, failureMessage));
      return new TestTheResultsWrapper<T>(op);
    }

    public static TestTheResultWrapper<T> TestForError<T>(
      this EntityScalarQueryOperation<T> op, string failureMessage = null) {

      WaitFor.IsCompleted(op);
      WaitFor.Callback(() => AssertNoQueryFailure(op, failureMessage));
      return new TestTheResultWrapper<T>(op);
    }

    public static TestTheResultWrapper<T> TestForError<T>(
      this EntityScalarQueryOperation<T> op, params Action<Exception>[] onErrorActions) {

      op.WireCompletedWithErrorActions(onErrorActions);
      return new TestTheResultWrapper<T>(op);
    }

    public static T TestForError<T>(this T op, string failureMessage = null)
      where T : BaseOperation {
      const string defaultFailureMessage = "Operation failed with exception {0}.";
      if (String.IsNullOrEmpty(failureMessage)) failureMessage = defaultFailureMessage;

      WaitFor.IsCompleted(op);
      WaitFor.Callback(() => AssertNoOperationFailure(op, failureMessage));

      return op;
    }

    private static void WireCompletedWithErrorActions(
      this BaseOperation op, IEnumerable<Action<Exception>> onErrorActions) {

      WaitFor.IsCompleted(op);
      WaitFor.Callback( 
        () => {
            if (!op.HasError) return;
            foreach (var onErrorAction in onErrorActions) {
              onErrorAction(op.Error);
            }                  
          });
    }

    internal static void AssertNoQueryFailure<T>(
      this EntityQueryOperation<T> op,
      string failureMessage = null) {

      const string defaultFailureMessage = "Query failed with exception {0}.";
      if (String.IsNullOrEmpty(failureMessage)) failureMessage = defaultFailureMessage;

      Assert.IsFalse(op.HasError, String.Format(failureMessage, op.Error));
    }

    internal static void AssertNoQueryFailure<T>(
      this EntityScalarQueryOperation<T> op,
      string failureMessage = null) {

      var queryTypeName = op.EntityQuery.GetType().Name;
      var defaultFailureMessage = "Scalar query [" + queryTypeName + "] failed with exception {0}.";
      if (String.IsNullOrEmpty(failureMessage)) failureMessage = defaultFailureMessage;

      Assert.IsFalse(op.HasError, String.Format(failureMessage, op.Error));
    }

    internal static void AssertNoOperationFailure(
      this BaseOperation op,
      string failureMessage = null) {

      const string defaultFailureMessage = "Operation failed with exception {0}.";
      if (String.IsNullOrEmpty(failureMessage)) failureMessage = defaultFailureMessage;

      Assert.IsFalse(op.HasError, String.Format(failureMessage, op.Error));
    }

    #region TestForError Fluent Interface Helpers

    public class TestTheResultsWrapper<T> {
      public TestTheResultsWrapper(EntityQueryOperation<T> op) {
        Operation = op;
      }

      public void TestTheResults(params Action<IEnumerable<T>>[] testActions) {
        Operation.TestTheResults(testActions);
      }

      internal EntityQueryOperation<T> Operation { get; private set; }

    }

    public class TestTheResultWrapper<T> {
      public TestTheResultWrapper(EntityScalarQueryOperation<T> op) {
        Operation = op;
      }

      public void TestTheResult(params Action<T>[] testActions) {
        Operation.TestTheResult(testActions);
      }

      internal EntityScalarQueryOperation<T> Operation { get; private set; }

    }

    #endregion

    public static void ShouldFail<T>(this BaseOperation op, string failureMessage = null)
      where T : Exception {

      WaitFor.IsCompleted(op);
      WaitFor.Callback(
        () => SoftwareApproach.TestingExtensions.Testing.ShouldThrowException<T>(
          () => { if (op.HasError) throw op.Error; }, failureMessage));
      WaitFor.TestComplete();
    }

  }

}