using System;
using System.Collections;
using System.Collections.Generic;
using IdeaBlade.EntityModel;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace IdeaBlade.TestFramework {

  public static class OperationTestTheResultsExtensions {

    /// <summary>
    /// Test the results of a successful <see cref="EntityQuery"/>
    /// for conformance with expectations.
    /// Raises TestComplete after tests.
    /// Throws <see cref="Assert.Fail(string)"/> if error or cancelled
    /// using <see cref="queryDescription"/> in the message.
    /// </summary>
    public static EntityQueryOperation TestTheResults(
      this EntityQueryOperation op,
      Action<IEnumerable> action,
      string queryDescription = null) {

      if (String.IsNullOrWhiteSpace(queryDescription)) {
        queryDescription = "EntityQueryOperation";
      }

      op.Completed += (s, args) => {
        ReportFailures(args, queryDescription);
        if (null != action) action(args.Results);
        WaitFor.TestComplete();
      };

      return op;

    }

    /// <summary>
    /// Test the results of a successful <see cref="EntityQuery{T}"/>
    /// for conformance with expectations.
    /// Raises TestComplete after tests.
    /// Throws <see cref="Assert.Fail(string)"/> if error or cancelled
    /// using <see cref="queryDescription"/> in the message.
    /// </summary>
    public static EntityQueryOperation<T> TestTheResults<T>(
      this EntityQueryOperation<T> op,
      Action<IEnumerable<T>> action,
      string queryDescription = null) {

      if (String.IsNullOrWhiteSpace(queryDescription)) {
        queryDescription = 
          String.Format("EntityQueryOperation<{0}", typeof(T).Name);
      }

      op.Completed += (s, args) => {
        ReportFailures(args, queryDescription);
        if (null != action) action(args.Results);
        WaitFor.TestComplete();
      };

      return op;

    }


    /// <summary>
    /// Test the result of a successful <see cref="EntityScalarQuery{T}"/>
    /// for conformance with expectations.
    /// Raises TestComplete after tests.
    /// Throws <see cref="Assert.Fail(string)"/> if error or cancelled
    /// using <see cref="queryDescription"/> in the message.
    /// </summary>
    public static EntityScalarQueryOperation<T> TestTheResult<T>(
      this EntityScalarQueryOperation<T> op,
      Action<T> action,
      string queryDescription = null) {

        if (String.IsNullOrWhiteSpace(queryDescription)) {
        queryDescription = 
          String.Format("EntityScalarQueryOperation<{0}", typeof(T).Name);
      }

      op.Completed += (s, args) => {
        ReportFailures(args, queryDescription);
        if (null != action) action(args.Result);
        WaitFor.TestComplete();
      };

      return op;

    }

    /// <summary>
    /// Test the result of a successful <see cref="EntityQuery"/>
    /// from a <see cref="EntityQueryCoroutineOperation"/>
    /// for conformance with expectations.
    /// Raises TestComplete after tests.
    /// Throws <see cref="Assert.Fail(string)"/> if error or cancelled
    /// using <see cref="queryDescription"/> in the message.
    /// </summary>
    public static EntityQueryCoroutineOperation TestTheResults(
      this EntityQueryCoroutineOperation op,
      Action<IEnumerable> action,
      string queryDescription = null) {

      if (String.IsNullOrWhiteSpace(queryDescription)) {
        queryDescription = op.TargetOperationName;
      }

      op.Completed += (s, args) => {
        ReportFailures(args, queryDescription);
        if (null != action) action(args.Results);
        WaitFor.TestComplete();
      };

      return op;
    }

    /// <summary>
    /// Test the result of a successful <see cref="EntityQuery{T}"/>
    /// from a <see cref="EntityQueryCoroutineOperation{T}"/>
    /// for conformance with expectations.
    /// Raises TestComplete after tests.
    /// Throws <see cref="Assert.Fail(string)"/> if error or cancelled
    /// using <see cref="queryDescription"/> in the message.
    /// </summary>
    public static EntityQueryCoroutineOperation<T> TestTheResults<T>(
      this EntityQueryCoroutineOperation<T> op,
      Action<IEnumerable<T>> action,
      string queryDescription = null) {

      if (String.IsNullOrWhiteSpace(queryDescription)) {
        queryDescription = op.TargetOperationName;
      }

      op.Completed += (s, args) => {
        ReportFailures(args, queryDescription);
        if (null != action) action(args.Results);
        WaitFor.TestComplete();
      };

      return op;
    }

    /// <summary>
    /// Test the result of a successful <see cref="EntityScalarQuery{T}"/>
    /// from a <see cref="EntityScalarQueryCoroutineOperation{T}"/>
    /// for conformance with expectations.
    /// Raises TestComplete after tests.
    /// Throws <see cref="Assert.Fail(string)"/> if error or cancelled
    /// using <see cref="queryDescription"/> in the message.
    /// </summary>
    public static EntityScalarQueryCoroutineOperation<T> TestTheResult<T>(
      this EntityScalarQueryCoroutineOperation<T> op,
      Action<T> action,
      string queryDescription = null) {

      if (String.IsNullOrWhiteSpace(queryDescription)) {
        queryDescription = op.TargetOperationName;
      }

      op.Completed += (s, args) => {
        ReportFailures(args, queryDescription);
        if (null != action) action(args.Result);
        WaitFor.TestComplete();
      };

      return op;
    }

    private static void ReportFailures<T>(T args, string operationDescription)
      where T : INotifyCompletedArgs {
      operationDescription = operationDescription ?? "Operation";
      if (args.Cancelled)
        Assert.Fail("Unexpected cancellation of " + operationDescription);
      if (null != args.Error) {
        args.IsErrorHandled = true; // we're handling by asserting failure
        Assert.Fail(operationDescription + " failed with exception: " + args.Error);          
      }
    }

    /// <summary>Expect a particular exception. Test for it.</summary>
    /// <typeparam name="TException">Type of expected exception</typeparam>
    /// <param name="op">The DevForce async <see cref="BaseOperation"/> object</param>
    /// <param name="operationDescription">
    /// Description of the operation being performed (optional).</param>
    /// <param name="errorMessageContains">
    /// Text fragment that must appear in the exception message (optiona).
    /// </param>
    /// <returns></returns>
    public static BaseOperation ExpectException<TException>(
      this BaseOperation op,
      string operationDescription = null,
      string errorMessageContains = null)
      where TException : Exception {

      op.Completed += (sender, args) => {

        if (String.IsNullOrWhiteSpace(operationDescription)) {
          operationDescription = op.GetType().Name;
        }

        var failMsg =
          String.Format("Expected \"{0}\" exception from {1} ",
                         typeof(TException).Name, operationDescription);

        if (args.HasError) {
          args.CheckExpectedException<TException>(failMsg, operationDescription, errorMessageContains);
        } else if (args.Cancelled) {
          Assert.Fail(failMsg + "but operation was cancelled instead.");
        } else {
          Assert.Fail(failMsg + "but there was no exception at all.");
        }
      };

      return op;
    }

    /// <summary>Expect a particular exception. Test for it.</summary>
    /// <typeparam name="TException">Type of expected exception</typeparam>
    /// <param name="op">The DevForce async <see cref="BaseOperation"/> object</param>
    /// <param name="operationDescription">
    /// Description of the operation being performed (optional).</param>
    /// <param name="errorMessageContains">
    /// Text fragment that must appear in the exception message (optiona).
    /// </param>
    /// <returns></returns>
    public static CoroutineOperationWrapper ExpectException<TException>(
      this CoroutineOperationWrapper op,
      string operationDescription = null,
      string errorMessageContains = null)
      where TException : Exception {

      op.WhenCompleted( args => {

        if (String.IsNullOrWhiteSpace(operationDescription)) {
          operationDescription = op.TargetOperationName;
        }

        var failMsg =
          String.Format("Expected \"{0}\" exception from {1} ",
                         typeof(TException).Name, operationDescription);

        if (args.Cancelled) {
          Assert.Fail(failMsg + "but operation was cancelled instead.");
        } else if (null != args.Error) {
          args.CheckExpectedException<TException>(failMsg, operationDescription, errorMessageContains);
        } else {
          Assert.Fail(failMsg + "but there was no exception at all.");
        }

      });

      return op;
    }


    private static void CheckExpectedException<TException>(
      this INotifyCompletedArgs args, 
      string failMsg, 
      string operationDescription,  
      string errorMessageContains ) {

      var err = args.Error;
      var errDescription = err.GetType() + " - " + err.Message;
      if (err is TException) {
        // got expected exception; now does it contain the expected error message fragment?
        if (!String.IsNullOrEmpty(errorMessageContains) &&
            !err.Message.Contains(errorMessageContains)) {

          failMsg = String.Format(
            "{0}with message containing \"{1}\" but actual exception was \"{2}\".",
            failMsg, errorMessageContains, errDescription);

          Assert.Fail(failMsg);
        } else {
          Console.WriteLine(operationDescription + " threw expected exception: " + errDescription);
          args.IsErrorHandled = true;
          WaitFor.TestComplete();
        }
      } else {
        Assert.Fail(failMsg + "but actual exception was \"" + errDescription + "\".");
      }
    }
  }

}