﻿using System.Linq;
using System.Text.RegularExpressions;
using IdeaBlade.Core;
using IdeaBlade.EntityModel;
using Microsoft.Silverlight.Testing;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using SimpleNorthwindModel;

// ReSharper disable InconsistentNaming
// ReSharper disable RedundantDefaultFieldInitializer
// ReSharper disable RedundantArgumentName

namespace WorkingWithEntities.PropertyInterception
{

    [TestClass]
    public class When_changing_customer_in_cache
    {
        [TestInitialize]
        public void TestInitialize()
        {
            manager = new EntityManager(shouldConnect: false);
            testCustomer = new Customer
            {
                CompanyName = "Acme",
                Country = "Freedonia",
            };
            manager.AddEntity(testCustomer);
        }

        /// <summary>Test TrimCompanyName attributed interceptor</summary>
        [TestMethod]
        [Tag("Intercept")]
        public void interceptor_trims_CompanyName()
        {
            const string newRawName = "   spaces  all   over  ";

            testCustomer.CompanyName = newRawName;

            var newTrimmedName = newRawName.Trim();
            Assert.AreEqual(newTrimmedName, testCustomer.CompanyName);
        }

        /// <summary>Test TrimCompanyName attributed interceptor with null</summary>
        [TestMethod]
        [Tag("Intercept")]
        public void interceptor_accepts_null_CompanyName()
        {
            testCustomer.CompanyName = null; // interceptor doesn't throw
            Assert.IsNull(testCustomer.CompanyName);
            //N.B.: It is invalid but it didn't throw
        }

        /// <summary>Test ContactName w/o runtime interceptor</summary>
        [TestMethod]
        [Tag("Intercept")]
        public void without_interceptor_ContactName_keeps_whitespace()
        {
            const string newRawContactName = "   spaces  all   over  ";

            // Ensure no property interceptor set for this property.
            RemoveContactNameInterceptor();

            testCustomer.ContactName = newRawContactName;

            Assert.IsTrue(testCustomer.ContactName.Contains(" "),
                "Space found in ContactName");
        }

        /// <summary>Test null ContactName with runtime interceptor</summary>
        [TestMethod]
        [Tag("Intercept")]
        public void interceptor_accepts_null_ContactName()
        {
            AddContactNameInterceptor();

            testCustomer.ContactName = null;
            Assert.IsNull(testCustomer.ContactName);
        }

        /// <summary>Test space-filled ContactName with runtime interceptor</summary>
        [TestMethod]
        [Tag("Intercept")]
        public void interceptor_strips_ContactName_whitespace()
        {
            AddContactNameInterceptor();

            const string newRawContactName = "   spaces  all   over  ";
            testCustomer.ContactName = newRawContactName;

            Assert.IsFalse(testCustomer.ContactName.Contains(" "),
                "Space found in ContactName");
        }

        protected EntityManager manager;
        protected Customer testCustomer;

        private static void AddContactNameInterceptor() {
          if (PropertyInterceptorManager.CurrentInstance
            .GetActions<DataEntityPropertySetInterceptorArgs<Customer, string>>(typeof(Customer), Customer.EntityPropertyNames.ContactName, PropertyInterceptorMode.BeforeSet)
            .Count > 0) return;

          var act = new PropertyInterceptorAction<
                        DataEntityPropertySetInterceptorArgs<Customer, string>>(
                  typeof(Customer),
                  Customer.EntityPropertyNames.ContactName,
                  PropertyInterceptorMode.BeforeSet,
                  args => {
                    if (null == args.Value) return;
                    args.Value = Regex.Replace(args.Value, @"\s", "");
                  }
              );

          PropertyInterceptorManager.CurrentInstance.AddAction(act);
        }

        private static void RemoveContactNameInterceptor() {
          var actions = PropertyInterceptorManager.CurrentInstance
            .GetActions<DataEntityPropertySetInterceptorArgs<Customer, string>>(typeof(Customer), Customer.EntityPropertyNames.ContactName, PropertyInterceptorMode.BeforeSet).ToList();
          actions.ForEach(a => PropertyInterceptorManager.CurrentInstance.RemoveAction(a));
        }
    }
}