﻿using System;
using System.ComponentModel;
using IdeaBlade.EntityModel;
using Microsoft.Silverlight.Testing;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using SimpleNorthwindModel;

// ReSharper disable InconsistentNaming
// ReSharper disable RedundantDefaultFieldInitializer
// ReSharper disable RedundantArgumentName

namespace WorkingWithEntities.UndoTests
{

    [TestClass]
    public class When_undoing_test_customer
    {
        [TestInitialize]
        public void TestInitialize()
        {
            manager = new EntityManager(shouldConnect: false);
            testCustomer = new Customer {CustomerID = Guid.NewGuid()};
            manager.AttachEntity(testCustomer);

            propertyChangedPropertyName = propertyChangedNotRaised;
            var inpc = ((INotifyPropertyChanged) testCustomer);
            inpc.PropertyChanged += 
                (s, e) => propertyChangedPropertyName = e.PropertyName;
        }
        
        [TestMethod]
        public void then_testCustomer_isUnchanged_at_start()
        {
            Assert.IsTrue(TestCustomerEntityState.IsUnchanged());            
        }

        [TestMethod]
        [Tag("Undo")]
        public void then_undo_unmodified_via_entity_raises_PropertyChanged_for_entity()
        {
            testCustomer.EntityAspect.RejectChanges();
            Assert.IsTrue(TestCustomerEntityState.IsUnchanged(),
                "should be unchanged");
            AssertPropertyChangedForEntireEntity();
        }

        [TestMethod]
        [Tag("Undo")]
        public void then_undo_unmodified_via_manager_does_nothing()
        {
            manager.RejectChanges();
            Assert.IsTrue(TestCustomerEntityState.IsUnchanged(),
                "should be unchanged");
            Assert.AreEqual(propertyChangedNotRaised, propertyChangedPropertyName,
                 "unexpected propChangedName == " + propertyChangedPropertyName);
        }

        [TestMethod]
        [Tag("Undo")]
        public void can_undo_changes_via_entity()
        {
            testCustomer.CompanyName = "Acme";
            Assert.IsTrue(TestCustomerEntityState.IsModified(), 
                "should be modified");
            testCustomer.EntityAspect.RejectChanges();
            Assert.IsTrue(TestCustomerEntityState.IsUnchanged(),
                "should be unchanged");
        }

        [TestMethod]
        [Tag("Undo")]
        public void can_undo_changes_via_manager()
        {
            testCustomer.CompanyName = "Acme";
            Assert.IsTrue(TestCustomerEntityState.IsModified(),
                "should be modified");
            manager.RejectChanges();
            Assert.IsTrue(TestCustomerEntityState.IsUnchanged(),
                "should be unchanged");
        }

        [TestMethod]
        [Tag("Undo")]
        public void then_undo_changes_via_entity_raises_PropertyChanged_for_entity()
        {
            testCustomer.CompanyName = "Acme"; // make a change
            testCustomer.EntityAspect.RejectChanges();
            AssertPropertyChangedForEntireEntity();
        }

        [TestMethod]
        [Tag("Undo")]
        public void then_undo_changes_via_manager_raises_PropertyChanged_for_entity()
        {
            testCustomer.CompanyName = "Acme"; // make a change
            manager.RejectChanges();
            AssertPropertyChangedForEntireEntity();
        }

        [TestMethod]
        [Tag("Undo")]
        public void then_undo_new_customer_via_entity_raises_PropertyChanged_for_entity()
        {
            testCustomer.EntityAspect.SetAdded(); // pretend is new
            testCustomer.EntityAspect.RejectChanges();
            AssertPropertyChangedForEntireEntity();
        }

        [TestMethod]
        [Tag("Undo")]
        public void then_undo_new_customer_via_manager_raises_PropertyChanged_for_entity()
        {
            testCustomer.EntityAspect.SetAdded(); // pretend is new
            manager.RejectChanges();
            AssertPropertyChangedForEntireEntity();
        }


        [TestMethod]
        [Tag("Undo")]
        public void then_undo_new_customer_detaches_it()
        {
            testCustomer.EntityAspect.SetAdded(); // pretend is new
            testCustomer.EntityAspect.RejectChanges();
            Assert.AreEqual(
                EntityState.Detached,
                testCustomer.EntityAspect.EntityState,
                "Expected detached after RejectChanges to added entity"
                );
        }

        #region Private helpers

        private EntityState TestCustomerEntityState
        {
            get { return testCustomer.EntityAspect.EntityState; }
        }

        private void AssertPropertyChangedForEntireEntity()
        {
            Assert.AreEqual(expectedPropChangedName, propertyChangedPropertyName,
                            "unexpected propChangedName == " + propertyChangedPropertyName);
        }
        private Customer testCustomer;
        private EntityManager manager;

        private string propertyChangedPropertyName;
        private const string propertyChangedNotRaised = "PropertyChanged not raised";

        // http://msdn.microsoft.com/en-us/library/system.componentmodel.inotifypropertychanged.propertychanged.aspx
        // "The PropertyChanged event can indicate all properties on the object have changed 
        // by using either null or String.Empty as the property name in the PropertyChangedEventArgs."
#if SILVERLIGHT
    // The PropertyName for a whole entity change is empty string in DF SL
        private readonly string expectedPropChangedName = String.Empty;
#else
        // The PropertyName for a whole entity change is null in DF .NET
        private const string expectedPropChangedName = null;
#endif

        #endregion

    }
}