Imports System.Collections.ObjectModel
Imports Microsoft.VisualStudio.TestTools.UnitTesting

Namespace IdeaBlade.TestFramework

  ''' <summary>
  ''' Base class for Microsoft Unit Test classes with asynchronous calls.
  ''' Provides a uniform approach to asynchronous testing in both
  ''' Full ("Desktop") .NET and Silverlight environments.
  ''' See its most important method, <see cref="DoItAsync"/>,
  ''' for an example of writing an asynch Test Method.
  ''' </summary>
  Partial Public Class DevForceTest

	Public Sub New()
	  WaitFor = New NestedWaitFor(Me)
	End Sub

	<ThreadStatic>
	Public Shared TestClassBase As DevForceTest

	<TestInitialize>
	Public Sub TestInitialize()
	  SetTestClassBase()
	  Context()
	End Sub

	''' <summary>
	''' Override to provide specific context for test methods of derived class.
	''' Extends this base class' MS Test [TestInitialize] method.
	''' </summary>
	Protected Overridable Sub Context()
	End Sub

	''' <summary>
	''' Perform the async action and block the test method thread
	''' until the action finishes.
	''' </summary>
	''' <param name="act">
	''' Void-returning delegate, anonymous delegates work fine too.
	''' </param>
	''' <remarks>
	''' Wrap an asynchronous action and its tests so that the Test Runner
	''' will hold the test method open until the async action completes.
	''' <para>
	''' Combines EnqueueCallback(Action) with WaitForTasksToFinish.
	''' </para>
	''' </remarks>
	''' <example>
	''' In this test which gets all customers asynchronously 
	''' the query is wrapped in a DoItAsync(...) call.
	''' <para>
	''' The parameter is an action that performs the query
	''' asynchronously and calls "TestComplete()" when done.
	''' </para><para>
	''' TestComplete tells the Test Runner that this test method
	''' has actually finished. Essential to call this method
	''' or else test Mmthod hangs indefinitely.
	''' It's a good idea to decorate your test method with 
	''' a <see cref="TimeoutAttribute"/> 
	''' to ensure that test method eventually closes.
	''' </para><para>
	''' The Test Runner automatically closes the test when the action throws
	''' an exception ... as it does when an assertion fails.
	''' </para>
	''' <![CDATA[
	''' [TestMethod, Asynchronous, Timeout(20000)]
	''' public void then_can_get_customers() {
	'''    DoItAsync(() =>
	''' 
	'''      Manager
	'''        .Customers
	'''        .ExecuteAsync()
	'''        .Completed += (sender, args) => {
	'''           args.HasError.ShouldBeFalse("Query failed with exception: " + args.Error); 
	'''           args.Results.ShouldNotBeEmpty("Should have customers"); 
	'''           TestComplete(); // <--- don't forget this!
	'''         }
	'''    );
	''' }
	''' ]]>
	''' </example>
	Protected Friend Sub DoItAsync(ByVal act As Action)
	  EnqueueCallback(act)
	  WaitForTasksToFinish()
	End Sub

	''' <summary>
	''' Add a Callback method into the test task queue. 
	''' Synonym for the EnqueueCallback(Action) method.
	''' </summary>
	''' <param name="testCallback">
	''' Void-returning delegate, anonymous delegates work fine too.
	''' </param>
	Protected Friend Sub WaitForIt(ByVal testCallback As Action)
	  EnqueueCallback(testCallback)
	End Sub

	Partial Private Sub SetTestClassBase()
	End Sub

  End Class
End Namespace