Imports IdeaBlade.EntityModel

Namespace IdeaBlade.TestFramework
  Public Module TestingExtensions

	''' <summary>
	''' Asserts that this <see cref="IEnumerable{T}"/> does not contain any items. 
	''' </summary>
    <System.Runtime.CompilerServices.Extension()> _
    Public Function ShouldBeEmpty(Of T)(ByVal source As IEnumerable(Of T), ByVal message As String, ByVal ParamArray parameters() As Object) As IEnumerable(Of T)
      If source.Any() Then
        AssertHelper.HandleFail("ShouldBeEmpty", message, parameters)
      End If
      Return source
    End Function

	''' <summary>
	''' Asserts that this <see cref="IEnumerable{T}"/> contains items. 
	''' </summary>
    <System.Runtime.CompilerServices.Extension()> _
    Public Function ShouldNotBeEmpty(Of T)(ByVal source As IEnumerable(Of T), ByVal message As String, ByVal ParamArray parameters() As Object) As IEnumerable(Of T)
      If Not source.Any() Then
        AssertHelper.HandleFail("ShouldNotBeEmpty", message, parameters)
      End If
      Return source
    End Function

	'----------------

	''' <summary>
	''' Asserts that this string contains a target string. 
	''' </summary>
    <System.Runtime.CompilerServices.Extension()> _
    Public Function ShouldContain(ByVal source As String, ByVal target As String, ByVal message As String, ByVal ParamArray parameters() As Object) As String
      If Not source.Contains(target) Then
        AssertHelper.HandleFail("ShouldContain", message, parameters)
      End If
      Return source
    End Function

	''' <summary>
	''' Asserts that this string does not contains a target string. 
	''' </summary>
    <System.Runtime.CompilerServices.Extension()> _
    Public Function ShouldNotContain(ByVal source As String, ByVal target As String, ByVal message As String, ByVal ParamArray parameters() As Object) As String
      If source.Contains(target) Then
        AssertHelper.HandleFail("ShouldNotContain", message, parameters)
      End If
      Return source
    End Function

	'----------------

	''' <summary>
	''' Asserts that this <see cref="IEnumerable{T}"/> contains a specific item. 
	''' </summary>
    <System.Runtime.CompilerServices.Extension()> _
    Public Function ShouldContain(Of T)(ByVal source As IEnumerable(Of T), ByVal item As T, ByVal message As String, ByVal ParamArray parameters() As Object) As IEnumerable(Of T)
      If Not source.Contains(item) Then
        AssertHelper.HandleFail("ShouldContain", message, parameters)
      End If
      Return source
    End Function

	''' <summary>
	''' Asserts that this <see cref="IEnumerable{T}"/> does not contain a specific item. 
	''' </summary>
    <System.Runtime.CompilerServices.Extension()> _
    Public Function ShouldNotContain(Of T)(ByVal source As IEnumerable(Of T), ByVal item As T, ByVal message As String, ByVal ParamArray parameters() As Object) As IEnumerable(Of T)
      If source.Contains(item) Then
        AssertHelper.HandleFail("ShouldNotContain", message, parameters)
      End If
      Return source
    End Function

	'----------------

	''' <summary>
	''' Asserts that this actual <see cref="IEnumerable{T}"/> 
	''' contains all of the items in the expected list. 
	''' </summary>
    <System.Runtime.CompilerServices.Extension()> _
    Public Function ShouldContainAllOf(Of T)(ByVal actual As IEnumerable(Of T), ByVal expected As IEnumerable(Of T), ByVal message As String, ByVal ParamArray parameters() As Object) As IEnumerable(Of T)
      If expected.Any(Function(exp) (Not actual.Contains(exp))) Then
        AssertHelper.HandleFail("ShouldContainAll", message, parameters)
      End If
      Return expected
    End Function

	''' <summary>
	''' Asserts that this actual <see cref="IEnumerable{T}"/> 
	''' does not contain any of the items in the expected list. 
	''' </summary>
    <System.Runtime.CompilerServices.Extension()> _
    Public Function ShouldNotContainAnyOf(Of T)(ByVal actual As IEnumerable(Of T), ByVal expected As IEnumerable(Of T), ByVal message As String, ByVal ParamArray parameters() As Object) As IEnumerable(Of T)
      If expected.Any(Function(exp) actual.Contains(exp)) Then
        AssertHelper.HandleFail("ShouldNotContainAnyOf", message, parameters)
      End If
      Return expected
    End Function

	'----------------

	''' <summary>
	''' Asserts that this entity is either null or the null entity. 
	''' </summary>
    <System.Runtime.CompilerServices.Extension()> _
    Public Function ShouldBeNullOrNullEntity(Of T As IEntity)(ByVal actual As T, ByVal message As String, ByVal ParamArray parameters() As Object) As T
      Return actual.AssertOnNullOrNullEntity(True, message, parameters)
    End Function

	''' <summary>
	''' Asserts that this entity is neither null nor the null entity. 
	''' </summary>
    <System.Runtime.CompilerServices.Extension()> _
    Public Function ShouldNotBeNullOrNullEntity(Of T As IEntity)(ByVal actual As T, ByVal message As String, ByVal ParamArray parameters() As Object) As T
      Return actual.AssertOnNullOrNullEntity(False, message, parameters)
    End Function

    <System.Runtime.CompilerServices.Extension()> _
    Private Function AssertOnNullOrNullEntity(Of T As IEntity)(ByVal actual As T, ByVal shouldBe As Boolean, ByVal message As String, ByVal ParamArray parameters() As Object) As T

      Dim isNull = Nothing Is actual OrElse actual.EntityAspect.IsNullEntity
      If shouldBe = isNull Then
        Return actual
      End If

      Const stateName As String = "null or null entity"
      Dim shouldText = (If(shouldBe, "Should be ", "Should not be ")) & stateName

      If 0 = parameters.Length Then
        message = String.Format("{0} of type '{1}'{2}.", shouldText, GetType(T).Name, message)
      End If
      Dim assertionName = shouldText.Replace(" ", String.Empty)
      AssertHelper.HandleFail(assertionName, message, parameters)

      Return actual
    End Function

	'========= ENTITYSTATE ASSERTIONS ==========

    <System.Runtime.CompilerServices.Extension()> _
    Private Function AssertOnEntityStateTest(Of T As IEntity)(ByVal actual As T, ByVal stateName As String, ByVal predicate As Func(Of Boolean), ByVal shouldBe As Boolean, ByVal message As String, ByVal ParamArray parameters() As Object) As T

      actual.ShouldNotBeNullOrNullEntity(Nothing)
      If shouldBe = predicate() Then
        Return actual
      End If

      Dim shouldText = (If(shouldBe, "Should be ", "Should not be ")) & stateName

      If 0 = parameters.Length Then
        message = String.Format("{0} but EntityState is '{1}'{2}.", shouldText, actual.EntityAspect.EntityState, message)
      End If
      Dim assertionName = shouldText.Replace(" ", String.Empty)
      AssertHelper.HandleFail(assertionName, message, parameters)

      Return actual
    End Function

	''' <summary>
	''' Asserts that this entity is Changed. 
	''' </summary>
    <System.Runtime.CompilerServices.Extension()> _
    Public Function ShouldBeChanged(Of T As IEntity)(ByVal actual As T, ByVal message As String, ByVal ParamArray parameters() As Object) As T

      Return actual.AssertOnEntityStateTest("Changed", Function() actual.EntityAspect.IsChanged, True, message, parameters)
    End Function

	''' <summary>
	''' Asserts that this entity is not Changed. 
	''' </summary>
    <System.Runtime.CompilerServices.Extension()> _
    Public Function ShouldNotBeChanged(Of T As IEntity)(ByVal actual As T, ByVal message As String, ByVal ParamArray parameters() As Object) As T

      Return actual.AssertOnEntityStateTest("Changed", Function() actual.EntityAspect.IsChanged, False, message, parameters)
    End Function

	''' <summary>
	''' Asserts that this entity is Unchanged. 
	''' </summary>
	''' <remarks>
	''' Should be equivalent to <see cref="ShouldNotBeChanged{T}(T, string , object[])"/>
	''' although implemented in terms of the <see cref="EntityState"/>.IsUnchanged() method.
	''' </remarks>
    <System.Runtime.CompilerServices.Extension()> _
    Public Function ShouldBeUnchanged(Of T As IEntity)(ByVal actual As T, ByVal message As String, ByVal ParamArray parameters() As Object) As T

      Return actual.AssertOnEntityStateTest("Unchanged", Function() actual.EntityAspect.EntityState.IsUnchanged(), True, message, parameters)
    End Function

	'----------------

	''' <summary>
	''' Asserts that this entity is Added. 
	''' </summary>
    <System.Runtime.CompilerServices.Extension()> _
    Public Function ShouldBeAdded(Of T As IEntity)(ByVal actual As T, ByVal message As String, ByVal ParamArray parameters() As Object) As T

      Return actual.AssertOnEntityStateTest("Added", Function() actual.EntityAspect.EntityState.IsAdded(), True, message, parameters)
    End Function

	''' <summary>
	''' Asserts that this entity is not Added. 
	''' </summary>
    <System.Runtime.CompilerServices.Extension()> _
    Public Function ShouldNotBeAdded(Of T As IEntity)(ByVal actual As T, ByVal message As String, ByVal ParamArray parameters() As Object) As T

      Return actual.AssertOnEntityStateTest("Added", Function() actual.EntityAspect.EntityState.IsAdded(), False, message, parameters)
    End Function

	'----------------

	''' <summary>
	''' Asserts that this entity is Deleted. 
	''' </summary>
    <System.Runtime.CompilerServices.Extension()> _
    Public Function ShouldBeDeleted(Of T As IEntity)(ByVal actual As T, ByVal message As String, ByVal ParamArray parameters() As Object) As T

      Return actual.AssertOnEntityStateTest("Deleted", Function() actual.EntityAspect.EntityState.IsDeleted(), True, message, parameters)
    End Function

	''' <summary>
	''' Asserts that this entity is not Deleted. 
	''' </summary>
    <System.Runtime.CompilerServices.Extension()> _
    Public Function ShouldNotBeDeleted(Of T As IEntity)(ByVal actual As T, ByVal message As String, ByVal ParamArray parameters() As Object) As T

      Return actual.AssertOnEntityStateTest("Deleted", Function() actual.EntityAspect.EntityState.IsDeleted(), False, message, parameters)
    End Function

	'----------------

	''' <summary>
	''' Asserts that this entity is Detached. 
	''' </summary>
    <System.Runtime.CompilerServices.Extension()> _
    Public Function ShouldBeDetached(Of T As IEntity)(ByVal actual As T, ByVal message As String, ByVal ParamArray parameters() As Object) As T

      Return actual.AssertOnEntityStateTest("Detached", Function() actual.EntityAspect.EntityState.IsDetached(), True, message, parameters)
    End Function

	''' <summary>
	''' Asserts that this entity is not Detached. 
	''' </summary>
    <System.Runtime.CompilerServices.Extension()> _
    Public Function ShouldNotBeDetached(Of T As IEntity)(ByVal actual As T, ByVal message As String, ByVal ParamArray parameters() As Object) As T

      Return actual.AssertOnEntityStateTest("Detached", Function() actual.EntityAspect.EntityState.IsDetached(), False, message, parameters)
    End Function
  End Module
End Namespace