﻿' ReSharper disable RedundantUsingDirective
Imports System.ComponentModel ' keep for SL
' ReSharper restore RedundantUsingDirective
Imports IdeaBlade.EntityModel
Imports Microsoft.Silverlight.Testing
Imports Microsoft.VisualStudio.TestTools.UnitTesting
Imports SimpleNorthwindModel

' ReSharper disable InconsistentNaming
' ReSharper disable RedundantDefaultFieldInitializer
' ReSharper disable RedundantArgumentName

Namespace WorkingWithEntities.EntityCacheTests
	<TestClass>
	Public Class EntityCacheTests
		<TestInitialize>
		Public Sub TestInitialize()
			manager = New EntityManager(shouldConnect:= False)
		End Sub

		<TestMethod, Tag("EntityCache")>
		Public Sub then_no_EntityGroups_at_start()
		   Assert.AreEqual(0, manager.GetEntityGroups().Count())
		End Sub

		<TestMethod, Tag("EntityCache")>
		Public Sub then_asking_for_EntitytGroup_creates_it()
			Assert.AreEqual(0, manager.GetEntityGroups().Count(), "should have no grps at start")

			' Merely asking for the Customer grp
			Dim custGrp = manager.GetEntityGroup(Of Customer)()

			Assert.AreEqual(1, manager.GetEntityGroups().Count(), "should have a group after asking for it")

			Assert.IsNotNull(custGrp, "the grp should be non-null")
			Assert.AreEqual(0, custGrp.Count(), "the grp should be empty")
		End Sub

		<TestMethod, Tag("EntityCache")>
		Public Sub then_have_a_Customer_EntityGroup_after_adding_Customer()
			manager.AddEntity(New Customer())

			Dim grps = manager.GetEntityGroups()
			Assert.AreEqual(1, grps.Count(), "should have exactly one group")

			Dim custGrp = manager.GetEntityGroups().FirstOrDefault()

			Assert.IsNotNull(custGrp, "should have a grp after adding customer")
			Assert.AreSame(GetType(Customer), custGrp.EntityType, "should have a Customer grp after adding customer")
			Assert.AreEqual(1, custGrp.Count(), "should have one Customer in grp")
		End Sub


		<TestMethod, Tag("EntityCache")>
		Public Sub then_manager_has_no_EntityGroups_after_clearing()
			manager.AddEntity(New Customer()) ' adds Customer EntityGroup
			manager.Clear() ' clears the cache

			' Cache forgets the groups it used to have after clearing
			Assert.AreEqual(0, manager.GetEntityGroups().Count(), "should have 1 grp after clear")
		End Sub

		<TestMethod, Tag("EntityCache")>
		Public Sub then_EntityGroup_has_deleted_customer()
			Dim cust = New Customer()
			manager.AttachEntity(cust)

			cust.EntityAspect.Delete()
			Dim custGrp = manager.GetEntityGroup(Of Customer)()

      Assert.IsTrue(custGrp.Entities.Contains(cust), "cust grp should have deleted customer")
		End Sub

		<TestMethod, Tag("EntityCache")>
		Public Sub then_EntityGroup_CurrentEntities_lacks_deleted_customer()
			Dim cust = New Customer()
			manager.AttachEntity(cust)

			cust.EntityAspect.Delete()
			Dim custGrp = manager.GetEntityGroup(Of Customer)()

			Assert.IsFalse(custGrp.CurrentEntities.Contains(cust), "CurrentEntities should not have deleted customer")
		End Sub

		<TestMethod, Tag("EntityCache")>
		Public Sub then_can_listen_to_EntityGroup_changes()
			' Listen for changes to cache
			Dim cacheChanges = New List(Of EntityAction)()
			Dim cacheChangeCount As Integer = 0
			AddHandler manager.EntityChanged, Sub(s, e) cacheChanges.Add(e.Action)

			' Listen for changes to customers in cache
			Dim custChanges = New List(Of EntityAction)()
			Dim custChangeCount As Integer = 0
			Dim custGrp = manager.GetEntityGroup(Of Customer)()
			AddHandler custGrp.EntityChanged, Sub(s, e) custChanges.Add(e.Action)

			' Customer changes
			Dim cust = New Customer()
			manager.AddEntity(cust)
			cacheChangeCount += 1
			custChangeCount += 1
			cust.CompanyName = "Acme"
			cacheChangeCount += 1
			custChangeCount += 1
			cust.CompanyName = "Beta"
			cacheChangeCount += 1
			custChangeCount += 1
			cust.EntityAspect.AcceptChanges()
			cacheChangeCount += 1
			custChangeCount += 1
			cust.EntityAspect.Delete()
			cacheChangeCount += 1
			custChangeCount += 1

			' Employee change
			Dim emp = New Employee With {.EmployeeID = 42}
			manager.AttachEntity(emp)
			cacheChangeCount += 1

			Assert.AreEqual(cacheChangeCount, cacheChanges.Count(), "not all cache changes were signaled")

			Assert.AreEqual(custChangeCount, custChanges.Count(), "not all cust changes were signaled")

			Assert.IsTrue(cacheChangeCount > custChangeCount, "should have more cache changes than cust changes")
		End Sub

		<TestMethod>
		Public Sub can_find_a_customer_in_cache_by_key()
			Dim cust1 = MakeUnchangedCustomer("Acme")
			Dim foundCust = manager.FindEntity(cust1.EntityAspect.EntityKey)
			Assert.AreSame(cust1, foundCust)
		End Sub


		<TestMethod>
		Public Sub can_find_a_deleted_customer_in_cache_by_key()
			Dim cust1 = MakeUnchangedCustomer("Acme")
			cust1.EntityAspect.Delete()
			Dim foundCust = manager.FindEntity(cust1.EntityAspect.EntityKey, includeDeleted:=True)
			Assert.AreSame(cust1, foundCust)
		End Sub

		<TestMethod>
		Public Sub can_find_a_customer_in_cache_by_query()
			Dim cust1 = MakeUnchangedCustomer("Acme")

			Dim foundCust = manager.FindEntities(EntityState.AllButDetached).OfType(Of Customer)().Where(Function(c) c.CompanyName = "Acme").FirstOrDefault()

			Assert.AreSame(cust1, foundCust)
		End Sub

		<TestMethod>
		Public Sub can_transfer_entity_via_ECS()
			Dim cust1 = MakeUnchangedCustomer("Acme")

			' Get EntityCacheState with all entities in the manager (just one)
			Dim ecs = manager.CacheStateManager.GetCacheState()

			' Create 2nd manager
			Dim manager2 = New EntityManager(shouldConnect:= False)

			' Prove that it is empty
			Dim entities = manager2.FindEntities(EntityState.AllButDetached).OfType(Of Object)()
			Assert.AreEqual(0, entities.Count(), "new manager2 should be empty")

			' "Restore" with the contents of the EntityCacheState from manager
			manager2.CacheStateManager.RestoreCacheState(ecs)

			' Prove that manager2 has a customer with same ID as cust1
			Dim foundCust = manager2.FindEntity(cust1.EntityAspect.EntityKey)

			' But the foundCust is not the same as cust1
			' because cust1 still belongs to manager1
			Assert.AreNotSame(cust1, foundCust)

		End Sub

		#Region "Test helpers"

		Protected Function MakeAddedCustomer(ByVal companyName As String) As Customer
			Dim cust = New Customer With {.CompanyName = companyName, .Country = "Freedonia"}
			manager.AddEntity(cust)
			Return cust
		End Function

		Protected Function MakeUnchangedCustomer(ByVal companyName As String) As Customer
			Dim cust = MakeAddedCustomer(companyName)
			cust.EntityAspect.AcceptChanges()
			Return cust
		End Function

		Protected manager As EntityManager
		Protected testCustomer As Customer

		#End Region


	End Class
End Namespace